<?php
/* Copyright start
  Copyright (C) 2008 - 2020 Fortinet Inc.
  All rights reserved.
  FORTINET CONFIDENTIAL & FORTINET PROPRIETARY SOURCE CODE
  Copyright end */

namespace Cybersponse\CoreBundle\Controller;

use Cybersponse\UtilsBundle\ARCUtils;
use Cybersponse\UtilsBundle\Utils;
use Dunglas\ApiBundle\JsonLd\Response;
use Cybersponse\CoreBundle\Api\ApiEvents;
use Dunglas\ApiBundle\Event\Events;
use Cybersponse\WorkflowBundle\Subscriber\DataEvent;
use Dunglas\ApiBundle\Api\ResourceInterface;
use Cybersponse\AuthorizationBundle\Permissions\PermissionsDictionaryBuilder;
use Symfony\Component\HttpKernel\Exception\HttpException;
use Symfony\Component\HttpFoundation\Request;
use Cybersponse\MetadataBundle\Entity\ModelMetadata;

class UpsertController extends AuthorizationResourceController
{
    /**
     * @param ResourceInterface $resource
     * @param array|object $data
     * @param int $status
     * @param array $headers
     * @param array $additionalContext
     * @param array $__selectFields
     * @return Response
     */
    protected function getSuccessResponse(
        ResourceInterface $resource,
        $data,
        $status = 200,
        array $headers = [],
        array $additionalContext = [],
        array $__selectFields = []
    ) {
        $normalized_data = $this->get('serializer')->normalize(
            $data,
            'json-ld',
            $resource->getNormalizationContext() + $additionalContext
        );
        $response = new Response(
            $normalized_data,
            $status,
            $headers
        );
        if ($status == Response::HTTP_CREATED) {
            $iri = $this->get('api.iri_converter')->getIriFromItem($data);
            $response->headers->set('Location', $iri);
        }
        return $response;
    }

    public function get_unique_resource_fields($request_data, $resource)
    {
        if ($request_data->{'__unique'}) {
            return $request_data->{'__unique'};
        }
        $proxyProvider = $this->get('cybersponse.core.provider.proxy');
        $cacheClient = $proxyProvider->getCache();
        $cacheName = 'constraintField_' . $resource->getModuleType();
        if ($cacheClient->cacheContains($cacheName)) {
            return  $cacheClient->getFromCache($cacheName);
        }
        $managerRegistry = $this->container->get('doctrine');
        $manager = $managerRegistry->getManagerForClass(ModelMetadata::class);
        $modelMetaDataRepository = $manager->getRepository(ModelMetadata::class);
        $modelMetaData = $modelMetaDataRepository->findOneBy(array('type' => $resource->getModuleType()));
        if (!$modelMetaData) {
            return [];
        }
        $uniueConstraint = count($modelMetaData->getUniqueConstraint()) ? $modelMetaData->getUniqueConstraint()[0] : [];
        $columns = [];
        foreach ($uniueConstraint as $key => $value) {
            $columns = isset($value['columns']) ? $value['columns'] : [];
        }
        $cacheClient->persistToCache($cacheName, $columns, $proxyProvider->getCacheLifetime());
        return $columns;
    }

    public function upsertAction(Request $request, $moduletype)
    {
        if (in_array($moduletype, self::DISALLOWED_MODULES)) {
            throw new HttpException(Response::HTTP_METHOD_NOT_ALLOWED, 'Upsert Operation for ' . $moduletype . ' modules is not supported.');
        }
        $request_data = json_decode($request->getContent());
        $resource = $this->identifyResourceByShortName($moduletype);
        $filterItems = $this->get_unique_resource_fields($request_data, $resource);
        $replaceIfExist = isset($request_data->{'__replace'}) ? $request_data->{'__replace'} : true;
        $requestBody = $request_data;
        ARCUtils::updateReferenceToIriAndCheckForRelation($requestBody);

        $utilsObj = new Utils();
        $objectData = json_decode(json_encode($requestBody), true);
        
        $data_filters = $utilsObj::prepare_filter_data($objectData, $filterItems);
        $object = [];
        if ($data_filters) {
            $object = $utilsObj->findWithWhere($this, $resource, $data_filters);
        }
        if (count($object) && !$replaceIfExist) {
            return $this->getSuccessResponse($resource, $object[0]);
        }
        if (count($object)) {
            $this->denyAccessUnlessGranted($resource, PermissionsDictionaryBuilder::PERM_CRUD_UPDATE);
            $this->get(self::EVENT_DISPATCHER)->dispatch(ApiEvents::PRE_API_UPDATE, new DataEvent($resource, null));
            $object = $object[0];
            $this->ValidateOwnerShip($request, $object);
            $object = $this->updateRecord($request, $resource, $object, $objectData, $requestBody);
            if (get_class($object) === "Symfony\Component\Validator\ConstraintViolationList") {
                return $this->getErrorResponse($object);
            }
            return $this->getSuccessResponse($resource, $object);
        } else {
            $this->denyAccessUnlessGranted($resource, PermissionsDictionaryBuilder::PERM_CRUD_CREATE);
            $this->get(self::EVENT_DISPATCHER)->dispatch(ApiEvents::PRE_API_CREATE, new DataEvent($resource, null));
            if (array_key_exists("__link", $requestBody) || array_key_exists("__unlink", $requestBody)) {
                $requestBody = ARCUtils::mergePostRequest($requestBody);
            }
            $object = $this->createObject($requestBody, $resource, $resource->getDenormalizationContext());
            $this->get(self::EVENT_DISPATCHER)->dispatch(Events::PRE_CREATE_VALIDATION, new DataEvent($resource, $object));
            $violations = $this->get('validator')->validate($object, null, $resource->getValidationGroups());
            if (0 === count($violations)) {
                $object = $this->createRecord($resource, $request, $object, $requestBody);
                return $this->getSuccessResponse($resource, $object, Response::HTTP_CREATED);
            }
            return $this->getErrorResponse($violations);
        }
    }

    public function bulkUpsertAction(Request $request, $moduletype)
    {
        $successObjects = array();
        $failureObjects = array();
        if (in_array($moduletype, self::DISALLOWED_MODULES)) {
            throw new HttpException(Response::HTTP_METHOD_NOT_ALLOWED, 'Upsert Operation for ' . $moduletype . ' modules is not supported.');
        }
        $resource = $this->identifyResourceByShortName($moduletype);
        $request_data = json_decode($request->getContent());
        $filterItems = $this->get_unique_resource_fields($request_data, $resource);
        $requestBodys = property_exists($request_data, "__data") ? $request_data->__data : $request_data;
        $replaceIfExist = isset($request_data->{'__replace'}) ? $request_data->{'__replace'} : true;
        $utilsObj = new Utils();
        foreach ($requestBodys as $requestBody) {
            ARCUtils::updateReferenceToIriAndCheckForRelation($requestBody);
            $objectData = json_decode(json_encode($requestBody), true);
            $data_filters = $utilsObj::prepare_filter_data($objectData, $filterItems);

            $object = [];
            if ($data_filters) {
                $object = $utilsObj->findWithWhere($this, $resource, $data_filters);
            }
            if (count($object) && !$replaceIfExist) {
                array_push($successObjects, $object[0]);
                continue;
            }
            if (count($object)) {
                $this->denyAccessUnlessGranted($resource, PermissionsDictionaryBuilder::PERM_CRUD_UPDATE);
                $this->get(self::EVENT_DISPATCHER)->dispatch(ApiEvents::PRE_API_UPDATE, new DataEvent($resource, null));
                try {
                    $object = $object[0];
                    $this->ValidateOwnerShip($request, $object);
                    $object = $this->updateRecord($request, $resource, $object, $objectData, $requestBody);
                    if (get_class($object) === "Symfony\Component\Validator\ConstraintViolationList") {
                        array_push($failureObjects, $this->get('serializer')->normalize($object, self::HYDRA_ERROR));
                        continue;
                    }
                    array_push($successObjects, $object);
                    continue;
                } catch (\Exception $exception) {
                    $returnData['data'] = $requestBody;
                    $returnData['exception'] = $this->get('serializer')->normalize($exception, self::HYDRA_ERROR);
                    array_push($failureObjects, $returnData);
                    continue;
                }
            } else {
                $this->denyAccessUnlessGranted($resource, PermissionsDictionaryBuilder::PERM_CRUD_CREATE);
                $this->get(self::EVENT_DISPATCHER)->dispatch(ApiEvents::PRE_API_CREATE, new DataEvent($resource, null));
                if (array_key_exists("__link", $requestBody) || array_key_exists("__unlink", $requestBody)) {
                    $requestBody = ARCUtils::mergePostRequest($requestBody);
                }
                $object = $this->createObject($requestBody, $resource, $resource->getDenormalizationContext());
                $this->get(self::EVENT_DISPATCHER)->dispatch(Events::PRE_CREATE_VALIDATION, new DataEvent($resource, $object));
                $violations = $this->get('validator')->validate($object, null, $resource->getValidationGroups());
                if (0 === count($violations)) {
                    $object = $this->createRecord($resource, $request, $object, $requestBody);
                    array_push($successObjects, $object);
                    continue;
                }
                $returnData['data'] = $requestBody;
                $returnData['exception'] = $this->get('serializer')->normalize($violations, self::HYDRA_ERROR);
                array_push($failureObjects, $returnData);
            }
        }
        if (!count($failureObjects)) {
            return $this->getSuccessResponse($resource, $successObjects);
        } elseif (count($failureObjects) && count($successObjects)) {
            $partialUpdatedObject = array();
            $partialUpdatedObject["failure"] = $failureObjects;
            $normalized_success_data = $this->get('serializer')->normalize(
                $successObjects,
                self::NORMALIZER_METHOD,
                $resource->getNormalizationContext()
            );
            $partialUpdatedObject["success"] = $normalized_success_data;
            return new Response($partialUpdatedObject, Response::HTTP_MULTI_STATUS);
        } elseif (!count($successObjects)) {
            return new Response($failureObjects, Response::HTTP_BAD_REQUEST);
        }
    }

    public function bulkUpsertWorkflowAction(Request $request) {
        $request_data = json_decode($request->getContent());
        $requestDatas = property_exists($request_data, "__data") ? $request_data->__data : $request_data;
        $objectDatas = json_decode(json_encode($requestDatas), true);
        $objectDatas = $this->upsertWorkflows($objectDatas);
        $resource = $this->identifyResourceByShortName('workflows');
        return $this->bulkUpsertPlaybook($request, $objectDatas, $resource);
    }

    public function bulkUpsertWorkflowCollectionsAction(Request $request) {
        $request_data = json_decode($request->getContent());
        $requestDatas = property_exists($request_data, "__data") ? $request_data->__data : $request_data;
        $objectDatas = json_decode(json_encode($requestDatas), true);
        $objectDatas = $this->upsertWorkflowCollections($objectDatas);
        $resource = $this->identifyResourceByShortName('workflow_collections');
        return $this->bulkUpsertPlaybook($request, $objectDatas, $resource);
    }

    public function bulkUpsertPlaybook(Request $request, $objectDatas, $resource)
    {   
        $successObjects = array();
        $failureObjects = array();
        $utilsObj = new Utils();
        foreach ($objectDatas as $objectData) {
            $requestBody = json_decode(json_encode($objectData));
            $data_filters = $utilsObj::prepare_filter_data($objectData, ['uuid']);
            $object = [];
            if ($data_filters) {
                $object = $utilsObj->findWithWhere($this, $resource, $data_filters);
            }

            if (count($object)) {
                $this->denyAccessUnlessGranted($resource, PermissionsDictionaryBuilder::PERM_CRUD_UPDATE);
                $this->get(self::EVENT_DISPATCHER)->dispatch(ApiEvents::PRE_API_UPDATE, new DataEvent($resource, null));
                try {
                    $object = $object[0];
                    $this->ValidateOwnerShip($request, $object);
                    $object = $this->updateRecord($request, $resource, $object, $objectData, $requestBody);
                    if (get_class($object) === "Symfony\Component\Validator\ConstraintViolationList") {
                        array_push($failureObjects, $this->get('serializer')->normalize($object, self::HYDRA_ERROR));
                        continue;
                    }
                    array_push($successObjects, $object);
                    continue;
                } catch (\Exception $exception) {
                    $returnData['data'] = $objectData;
                    $returnData['exception'] = $this->get('serializer')->normalize($exception, self::HYDRA_ERROR);
                    array_push($failureObjects, $returnData);
                    continue;
                }
            } else {
                $this->denyAccessUnlessGranted($resource, PermissionsDictionaryBuilder::PERM_CRUD_CREATE);
                $this->get(self::EVENT_DISPATCHER)->dispatch(ApiEvents::PRE_API_CREATE, new DataEvent($resource, null));
                $object = $this->createObject($requestBody, $resource, $resource->getDenormalizationContext());
                $this->get(self::EVENT_DISPATCHER)->dispatch(Events::PRE_CREATE_VALIDATION, new DataEvent($resource, $object));
                $violations = $this->get('validator')->validate($object, null, $resource->getValidationGroups());
                if (0 === count($violations)) {
                    $object = $this->createRecord($resource, $request, $object, $requestBody);
                    array_push($successObjects, $object);
                    continue;
                }
                $returnData['data'] = $requestBody;
                $returnData['exception'] = $this->get('serializer')->normalize($violations, self::HYDRA_ERROR);
                array_push($failureObjects, $returnData);
            }
        }
        if (!count($failureObjects)) {
            return $this->getSuccessResponse($resource, $successObjects);
        } elseif (count($failureObjects) && count($successObjects)) {
            $partialUpdatedObject = array();
            $partialUpdatedObject["failure"] = $failureObjects;
            $normalized_success_data = $this->get('serializer')->normalize(
                $successObjects,
                self::NORMALIZER_METHOD,
                $resource->getNormalizationContext()
            );
            $partialUpdatedObject["success"] = $normalized_success_data;
            return new Response($partialUpdatedObject, Response::HTTP_MULTI_STATUS);
        } elseif (!count($successObjects)) {
            return new Response($failureObjects, Response::HTTP_BAD_REQUEST);
        }
    }

    public function modify_workflow_steps_routes($existingItems, $newItems, $prefix) {
        foreach($newItems as $index=>$newitem) {
            foreach($existingItems as $existingItem) {
                if ($existingItem->getUuid() == $newitem['uuid']) {
                    $newItems[$index]['@id'] = $prefix . $newitem['uuid'];
                    break;
                }
            }
        }
        return $newItems;
    }

    public function upsertWorkflowCollections($workflowCollections) {
        foreach($workflowCollections as $index=>$workflowCollection) {
            $workflows = $workflowCollection['workflows'];
            $workflowCollections[$index]['workflows'] = $this->upsertWorkflows($workflows);
        }
        return $workflowCollections;

    }


    public function upsertWorkflows($workflows) {
        $resource = $this->identifyResourceByShortName('workflows');
        $utilsObj = new Utils();
        foreach($workflows as $index=>$workflow) {
            $data_filters = $utilsObj::prepare_filter_data($workflow, ['uuid']);
            $object = $utilsObj->findWithWhere($this, $resource, $data_filters);

            if (count($object)) {
                $workflow['@id'] = '/api/3/workflows/' . $workflow['uuid'];
                $workflow['steps'] =  $this->modify_workflow_steps_routes($object[0]->getSteps(), $workflow['steps'], '/api/3/workflow_steps/');
                $workflow['routes'] = $this->modify_workflow_steps_routes($object[0]->getRoutes(), $workflow['routes'], '/api/3/workflow_routes/');
            }
            
            $workflows[$index] = $workflow;
        }
        return $workflows;

    }
}
